package de.fhhannover.inform.trust.ifmapj;

import java.io.IOException;

import javax.net.ssl.KeyManager;
import javax.net.ssl.SSLContext;
import javax.net.ssl.TrustManager;

import de.fhhannover.inform.trust.ifmapj.channel.SSRC;
import de.fhhannover.inform.trust.ifmapj.channel.SsrcImpl;
import de.fhhannover.inform.trust.ifmapj.exception.InitializationException;
import de.fhhannover.inform.trust.ifmapj.identifier.Identifier;
import de.fhhannover.inform.trust.ifmapj.identifier.IdentifierFactory;
import de.fhhannover.inform.trust.ifmapj.identifier.Identifiers;
import de.fhhannover.inform.trust.ifmapj.messages.RequestFactory;
import de.fhhannover.inform.trust.ifmapj.messages.Requests;
import de.fhhannover.inform.trust.ifmapj.metadata.StandardIfmapMetadataFactory;
import de.fhhannover.inform.trust.ifmapj.metadata.StandardIfmapMetadataFactoryImpl;


/**
 * Entry class to do IF-MAP 2.0 communication using IfmapJ.
 * 
 * @author aw
 *
 */
@SuppressWarnings("deprecation")
public class IfmapJ {
	
	/**
	 * Create a new {@link SSRC} object to operate on the given url
	 * using basic authentication.
	 * 
	 * @param url the URL to connect to
	 * @param user basic authentication user
	 * @param pass basic authentication password
	 * @param tms TrustManager instances to initialize the {@link SSLContext} with.
	 * @return a new {@link SSRC} that uses basic authentication
	 * @throws IOException 
	 */
	public static SSRC createSSRC(String url, String user, String pass, TrustManager[] tms)
			throws InitializationException {
		return new SsrcImpl(url, user, pass, tms);
	}
	
	/**
	 * Create a new {@link SSRC} object to operate on the given URL
	 * using certificate-based authentication.
	 * 
	 * The keystore and truststore to be used have to be set using
	 * the {@link System#setProperty(String, String)} method using
	 * the keys javax.net.ssl.keyStore, and javax.net.ssl.trustStore
	 * respectively.
	 * 
	 * @param url the URL to connect to
	 * @param kms TrustManager instances to initialize the {@link SSLContext} with.
	 * @param tms KeyManager instances to initialize the {@link SSLContext} with.
	 * @return a new {@link SSRC} that uses certificate-based authentication
	 * @throws IOException 
	 */
	public static SSRC createSSRC(String url, KeyManager[] kms, TrustManager[] tms)
			throws InitializationException {
		return new SsrcImpl(url, kms, tms);
	}
	
	/**
	 * Create a new {@link RequestFactory} object to create IF-MAP requests
	 * to be used with the methods provided by the {@link SSRC} class.
	 * 
	 * @return a new {@link RequestFactory} that is used to create IF-MAP requests
	 * @deprecated
	 * Use {@link Requests} directly.
	 */
	public static RequestFactory createRequestFactory() {
		return Requests.getRequestFactory();
	}
	
	/**
	 * Create a {@link IdentifierFactory} object to create different types
	 * of {@link Identifier} implementations.
	 * 
	 * @return a new {@link IdentifierFactory} that is used to create IF-MAP identifiers
	 * @deprecated
	 * Use {@link Identifiers} directly.
	 */
	public static IdentifierFactory createIdentifierFactory() {
		return Identifiers.getIdentifierFactory();
	}
	
	/**
	 * Crate a new {@link StandardIfmapMetadataFactory} object to create
	 * standard metadata defined by IF-MAP 2.0.
	 * 
	 * @return a new {@link StandardIfmapMetadataFactory} to create metadata
	 */
	public static StandardIfmapMetadataFactory createStandardMetadataFactory() {
		return new StandardIfmapMetadataFactoryImpl();
	}
}
