package de.fhhannover.inform.trust.ifmapj.channel;

import de.fhhannover.inform.trust.ifmapj.exception.CommunicationException;
import de.fhhannover.inform.trust.ifmapj.exception.IfmapErrorResult;
import de.fhhannover.inform.trust.ifmapj.exception.IfmapException;
import de.fhhannover.inform.trust.ifmapj.messages.Request;
import de.fhhannover.inform.trust.ifmapj.messages.Result;

/**
 * Interface for an IF-MAP communication channel
 * 
 * @author aw
 *
 */
public interface IfmapChannel {
	
	/**
	 * @return the current session-id for this channel, either set by
	 * a call to {@link SSRC#newSession()} or {@link #setSessionId(String)},
	 * or null if none of these calls was made.
	 */
	public String getSessionId();
	
	/**
	 * Set the session-id to be used for this channel. Be aware, calling this
	 * on an {@link ARC} will change the {@link SSRC} object attached to the
	 * {@link ARC}. Calling it on an {@link SSRC} will also change the attached
	 * {@link ARC} objects for the {@link SSRC}.
	 */
	public void setSessionId(String sessionId);
	
	/**
	 * @return the current ifmap-publisher-id for this channel, either set by
	 * a call to {@link SSRC#newSession()} or {@link #setPublisherId(String)},
	 * or null if none of these calls was made.
	 */
	public String getPublisherId();
	
	/**
	 * Set the ifmap-publisher-id for this channel. Be aware, calling this
	 * on an {@link ARC} will change the {@link SSRC} object attached to the
	 * {@link ARC}. Calling it on an {@link SSRC} will also change the attached
	 * {@link ARC} objects for the {@link SSRC}.
	 * 
	 * This call will only result in having {@link #getPublisherId()} return
	 * the set value. Requests won't be influenced.
	 */
	public void setPublisherId(String publisherId);
	
	/**
	 * @return the current max-poll-result-size for this channel, either set by
	 * a call to {@link SSRC#newSession()} or {@link #setMaxPollResSize(Integer)},
	 * or null if none of these calls was made.
	 */
	public Integer getMaxPollResSize();

	/**
	 * Set the max-poll-result-size for this channel. Be aware, calling this
	 * on an {@link ARC} will change the {@link SSRC} object attached to the
	 * {@link ARC}. Calling it on an {@link SSRC} will also change the attached
	 * {@link ARC} objects for the {@link SSRC}.
	 * 
	 * This call will only result in having {@link #getMaxPollResSize()} return
	 * the set value. Requests won't be influenced.
	 */
	public void setMaxPollResSize(Integer mprs);
	
	/**
	 * Closes the underlying TCP connection.
	 * 
	 * @throws CommunicationException
	 */
	public void closeTcpConnection() throws CommunicationException;
	
	/**
	 * Specifies whether GZIP compression should be used.
	 * 
	 * @param useGzip true if GZIP should be used
	 */
	public void setGzip(boolean useGzip);
	
	/**
	 * Indicates whether GZIP compression is used by this channel
	 * 
	 * @return true, if GZIP is used
	 */
	public boolean usesGzip();
	
	/**
	 * Generic Request Interface
	 * 
	 * @param req
	 * @return
	 * @throws IfmapErrorResult
	 * @throws IfmapException
	 */
	public Result genericRequest(Request req) throws IfmapErrorResult, IfmapException;
	public Result genericRequestWithSessionId(Request req) throws IfmapErrorResult, IfmapException;
}
