package de.fhhannover.inform.trust.ifmapj.channel;

import javax.net.ssl.KeyManager;
import javax.net.ssl.TrustManager;

import de.fhhannover.inform.trust.ifmapj.exception.CommunicationException;
import de.fhhannover.inform.trust.ifmapj.exception.IfmapErrorResult;
import de.fhhannover.inform.trust.ifmapj.exception.IfmapException;
import de.fhhannover.inform.trust.ifmapj.exception.InitializationException;
import de.fhhannover.inform.trust.ifmapj.messages.EndSessionRequest;
import de.fhhannover.inform.trust.ifmapj.messages.NewSessionRequest;
import de.fhhannover.inform.trust.ifmapj.messages.NewSessionResult;
import de.fhhannover.inform.trust.ifmapj.messages.PublishRequest;
import de.fhhannover.inform.trust.ifmapj.messages.PurgePublisherRequest;
import de.fhhannover.inform.trust.ifmapj.messages.RenewSessionRequest;
import de.fhhannover.inform.trust.ifmapj.messages.Requests;
import de.fhhannover.inform.trust.ifmapj.messages.Result;
import de.fhhannover.inform.trust.ifmapj.messages.SearchRequest;
import de.fhhannover.inform.trust.ifmapj.messages.SearchResult;
import de.fhhannover.inform.trust.ifmapj.messages.SubscribeRequest;

public class SsrcImpl extends AbstractChannel implements SSRC {
	
	/**
	 * the session-id received after the last call to {@link #newSession()}
	 */
	private String mSessionId;
	
	/**
	 * the publisher-id received after the last call to {@link #newSession()}
	 */
	private String mPublisherId;

	/**
	 * the max-poll-result-size received after the last call to {@link #newSession()},
	 * if any
	 */
	private Integer mMaxPollResultSize;
	
	/**
	 * KeyManager instances to initialize the SSL Context.
	 */
	private KeyManager[] mKeyManagers;
	
	/**
	 * TrustManager instances to initialize the SSL Context.
	 */
	private TrustManager[] mTrustManagers;

	/**
	 * Basic-Auth constructor.
	 * 
	 * @param url
	 * @param user
	 * @param pass
	 * @param tms
	 * @throws InitializationException
	 */
	public SsrcImpl(String url, String user, String pass, TrustManager[] tms)
			throws InitializationException {
		super(url, user, pass, tms);
		mTrustManagers = tms;
	}

	/**
	 * Certificate-based constructor.
	 * 
	 * @param url
	 * @param kms
	 * @param tms
	 * @throws InitializationException
	 */
	public SsrcImpl(String url, KeyManager[] kms, TrustManager[] tms)
			throws InitializationException {
		super(url, kms, tms);
		mKeyManagers = kms;
		mTrustManagers = tms;
	}
	
	@Override
	public void newSession(Integer maxPollResSize)
			throws IfmapErrorResult, IfmapException {
		
		NewSessionRequest nsreq = Requests.createNewSessionReq();
		
		nsreq.setMaxPollResultSize(maxPollResSize);
		
		Result res = genericRequest(nsreq);
		
		if (!(res instanceof NewSessionResult))
			throw new RuntimeException("wrong result-type");
		
		NewSessionResult nsres = (NewSessionResult) res;
			
		// reset session information
		mSessionId = null;
		mMaxPollResultSize = null;
		mPublisherId = null;
		mSessionId = nsres.getSessionId();
		mPublisherId = nsres.getPublisherId();
		
		if (maxPollResSize != null) {
			mMaxPollResultSize = nsres.getMaxPollResultSize();
			if (mMaxPollResultSize == null)
				throw new CommunicationException("no max-poll-result-size " +
						"in newSession result");
		}
	}


	@Override
	public void newSession() throws IfmapErrorResult, IfmapException {
		newSession(null);
	}

	@Override
	public void endSession() throws IfmapErrorResult, IfmapException {
		EndSessionRequest esr = Requests.createEndSessionReq();
		genericRequestWithSessionId(esr);
			
		mPublisherId = null;
		mSessionId = null;
		mMaxPollResultSize = null;
	}

	@Override
	public void renewSession() throws IfmapErrorResult, IfmapException {
		RenewSessionRequest rsr = Requests.createRenewSessionReq();
		genericRequestWithSessionId(rsr);
	}
		
	
	@Override
	public void purgePublisher() throws IfmapErrorResult, IfmapException {
		purgePublisher(getPublisherId());
	}

	@Override
	public void purgePublisher(String publisherId) 
			throws IfmapErrorResult, IfmapException {
		PurgePublisherRequest ppr = Requests.createPurgePublisherReq();
		ppr.setPublisherId(publisherId);
		genericRequestWithSessionId(ppr);
	}

	@Override
	public void publish(PublishRequest pr) throws IfmapErrorResult, IfmapException {
		if (pr == null)
			throw new NullPointerException();
	
		genericRequestWithSessionId(pr);
	}

	@Override
	public void subscribe(SubscribeRequest sr) throws IfmapErrorResult, IfmapException {
		if (sr == null)
			throw new NullPointerException();
		
		genericRequestWithSessionId(sr);
	}

	@Override
	public SearchResult search(SearchRequest sr) throws IfmapErrorResult, IfmapException {
		if (sr == null)
			throw new NullPointerException();
		
		Result res = genericRequestWithSessionId(sr);
		
		if (!(res instanceof SearchResult))
				throw new RuntimeException("search returns no SearchResult?");
		
		return (SearchResult)res;
	}

	@Override
	public ARC getArc() throws InitializationException {
		ARC ret = null;
		
		if (isBasicAuth())
			ret = new ArcImpl(this, getUrl(), getUser(), getPassword(), mTrustManagers);
		else
			ret = new ArcImpl(this, getUrl(), mKeyManagers, mTrustManagers);
		
		if (usesGzip())
			ret.setGzip(true);
		
		return ret;
	}

	@Override
	public String getSessionId() {
		return mSessionId;
	}

	@Override
	public String getPublisherId() {
		return mPublisherId;
	}

	@Override
	public Integer getMaxPollResSize() {
		return mMaxPollResultSize;
	}

	@Override
	public void setSessionId(String sessionId) {
		mSessionId = sessionId;
	}

	@Override
	public void setPublisherId(String publisherId) {
		mPublisherId = publisherId;
	}

	@Override
	public void setMaxPollResSize(Integer mprs) {
		mMaxPollResultSize = mprs;
	}
}
