package de.fhhannover.inform.trust.ifmapj.messages;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import de.fhhannover.inform.trust.ifmapj.exception.IfmapErrorResult;
import de.fhhannover.inform.trust.ifmapj.exception.MarshalException;
import de.fhhannover.inform.trust.ifmapj.exception.UnmarshalException;

/**
 * Provides functionality to generically handle {@link Request} instances and
 * response parsing.
 * 
 * @author aw
 * @since 0.1.4
 */
public interface RequestHandler<T extends Request> {
	
	/**
	 * Marshals a given {@link Request} instance to XML form and returns the
	 * {@link Element} to be attached to the soap:Body element.
	 * 
	 * @param req the {@link Request} instance to be marshalled.
	 * @param doc a {@link Document} instance to be used to create new
	 *        {@link Element} instances.
	 * @return an {@link Element} instance to be attached to the soap:Body.
	 * @throws MarshalException
	 */
	/**
	 * @throws MarshalException
	 */
	Element toElement(Request req, Document doc) throws MarshalException;
	
	/**
	 * Unmarshals a given response in XML form to a {@link Result} instance.
	 * The given {@link Element} instance is the ifmap:response element. It
	 * might have an errorResult element attached. An implementation has to
	 * check for this condition and throw an appropriate
	 * {@link IfmapErrorResult}.
	 * 
	 * @see {@link Requests.Helpers#getResponseContentErrorCheck(Element)}.
	 * 
	 * @param response the {@link Element} under <ifmap:response>
	 * @return an appropriate {@link Result} implementation or null if there is
	 *         no real result, but everything was good.
	 * @throws UnmarshalException
	 * @throws {@link IfmapErrorResult} if an errorResult was found
	 */
	Result fromElement(Element response) throws UnmarshalException, IfmapErrorResult;
		
	/**
	 * @return the {@link Class} object associated with the {@link Request}
	 *         instance this {@link RequestHandler} is able to cope with.
	 */
	Class<T> handles();
}
