package util;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import de.fhhannover.inform.trust.ifmapj.binding.IfmapStrings;
import de.fhhannover.inform.trust.ifmapj.exception.MarshalException;
import de.fhhannover.inform.trust.ifmapj.log.IfmapJLog;
import de.fhhannover.inform.trust.ifmapj.messages.NamespaceDeclarationHolder;

/**
 * Some helpers to handle org.w3c.* stuff...
 * 
 * @since 0.1.4
 * @author aw
 */
public class DomHelpers {
	
	private static final DocumentBuilderFactory sDocumentBuilderFactory;
	private static final DocumentBuilder sDocumentBuilder;
	private static final TransformerFactory sTransformerFactory;
	
	static {
		sTransformerFactory = TransformerFactory.newInstance();
		sDocumentBuilderFactory = DocumentBuilderFactory.newInstance();
		sDocumentBuilderFactory.setNamespaceAware(true);
		sDocumentBuilder = newDocumentBuilder();
	}
	
	public static final DocumentBuilder newDocumentBuilder() {
		try {
			return sDocumentBuilderFactory.newDocumentBuilder();
		} catch (ParserConfigurationException e) {
			IfmapJLog.error(e.getMessage());
			throw new RuntimeException(e.getMessage());
		}
	}
	
	public static final Element createNonNsElement(Document doc, String name) {
		return doc.createElementNS(null, name);
	}
	
	public static final void addAttribute(Element el, String name, String val) {
		el.setAttributeNS(null, name, val);
	}

	public static final void addXmlNamespaceDeclarations(NamespaceDeclarationHolder nsh,
			Element to) {
		
		for (Pair<String, String> nsDecl : nsh.getNamespaceDeclarations()) {
			String name = "xmlns";
			String value;
			if (nsDecl.first.length() > 0)
				name += ":" + nsDecl.first;
			value = nsDecl.second;
			to.setAttributeNS("http://www.w3.org/2000/xmlns/", name, value);
		}
	}

	public static final String makeRequestFQName(String name) {
		return IfmapStrings.BASE_PREFIX + ":" + name;
	}

	public static final boolean elementMatches(Element el, String elname) {
		return elementMatches(el, elname, IfmapStrings.NO_URI);
	}
	
	public static final boolean elementMatches(Element e, String name, String uri) {
		// i'm sorry
		return (name.equals(e.getLocalName())
				&& ((uri == null && e.getNamespaceURI() == null)
					|| (uri != null && uri.equals(e.getNamespaceURI()))));
	}
	
	/**
	 * Create a new list containing only the child nodes of e of type
	 * {@link Element}.
	 * @param e
	 * @return
	 */
	public static final List<Element> getChildElements(Element e) {
		
		List<Element> ret = new ArrayList<Element>();
		NodeList children = e.getChildNodes();
		
		for (int i = 0; i < children.getLength(); i++) {
			Node child = children.item(i);
			if (child.getNodeType()  == Node.ELEMENT_NODE)
				ret.add((Element)child);
		}
		
		return ret;
	}
	
	public static final Element findElementInChildren(Node node, String name,
			String uri) {
		NodeList children = node.getChildNodes();
		Element ret = null;
		for (int i = 0; i < children.getLength(); i++) {
			Node n = children.item(i);
			
			// our interest is in element nodes only
			if (!(n.getNodeType() == Node.ELEMENT_NODE))
				continue;
				
			if (elementMatches((Element)n, name, uri)) {
				ret = (Element)n;
				break;
			}
		}
		return ret;
	}

	/**
	 * Makes a deep-copy of an {@link Element} instance and appends it to a
	 * new {@link Document} instance as child.
	 * 
	 * @param child the {@link Element} instnace to be deep-copied
	 * @return a new {@link Document} instance with a deep-copied child appended
	 */
	public static final Document deepCopy(Element child) {
		Document mdDoc = sDocumentBuilder.newDocument();
		Element el = (Element) mdDoc.importNode(child, true);
		mdDoc.appendChild(el);
		return mdDoc;
	}
	
	
	/**
	 * Marshal a {@link Document} to {@link InputStream}.
	 * 
	 * @param doc
	 * @return
	 * @throws MarshalException
	 */
	public static final InputStream toInputStream(Document doc) throws MarshalException {
		
		Transformer trans;
		
		try {
			trans = sTransformerFactory.newTransformer();
		} catch (TransformerConfigurationException e) {
			IfmapJLog.error("Oh oh.... [" + e.getMessage() + "]");
			throw new MarshalException(e.getMessage());
		}
		
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		Result result = new StreamResult(baos);
		byte[] marshalled = null;
		
		doc.setXmlStandalone(true);
		Source source = new DOMSource(doc);

		
		try {
			trans.transform(source, result);
		} catch (TransformerException e) {
			IfmapJLog.error("Oh oh.... [" + e.getMessage() + "]");
			throw new MarshalException(e.getMessage());
		}
		try {
			baos.flush();
		} catch (IOException e) {
			IfmapJLog.error("Oh oh.... [" + e.getMessage() + "]");
			throw new MarshalException(e.getMessage());
		}
	
		marshalled = baos.toByteArray();
		
		return new ByteArrayInputStream(marshalled);
	}
}
