package de.fhhannover.inform.trust.ifmapj.identifier;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;

import javax.xml.parsers.DocumentBuilder;

import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import de.fhhannover.inform.trust.ifmapj.exception.MarshalException;
import de.fhhannover.inform.trust.ifmapj.exception.UnmarshalException;

import util.DomHelpers;

public class AccessRequestHandlerTest {
	
	private static DocumentBuilder sDocBuilder = DomHelpers.newDocumentBuilder();
	private static AccessRequestHandler sAccessRequestHandler = new AccessRequestHandler();
	
	@Test(expected=MarshalException.class)
	public void testToElementWithNullName() throws MarshalException {
		Document doc = sDocBuilder.newDocument();
		AccessRequest ar = Identifiers.createAr(null);
	
		// throws
		Element res = sAccessRequestHandler.toElement(ar, doc);
		assertNull(res);
	}

	@Test(expected=MarshalException.class)
	public void testToElementWithEmptyName() throws MarshalException {
		Document doc = sDocBuilder.newDocument();
		AccessRequest ar = Identifiers.createAr("");
	
		// throws
		Element res = sAccessRequestHandler.toElement(ar, doc);
		assertNull(res);
	}

	/**
	 * This shouldn't show the administrative-domain
	 * @throws MarshalException 
	 */
	@Test
	public void testToElementWithEmptyAd() throws MarshalException {
		Document doc = sDocBuilder.newDocument();
		AccessRequest ar = Identifiers.createAr("AR100", "");
		
		Element res = sAccessRequestHandler.toElement(ar, doc);
		assertNotNull(res);
		assertEquals("access-request", res.getLocalName());
		assertNull(res.getPrefix());
		assertNull(res.getNamespaceURI());
		assertNotNull(res.getAttributeNode("name"));
		assertEquals("AR100", res.getAttribute("name"));
		
		assertNull(res.getAttributeNode("administrative-domain"));
		
		assertEquals(0, res.getChildNodes().getLength());
		assertEquals(1, res.getAttributes().getLength());
	}
		
	@Test
	public void testToElementWithName() throws MarshalException {
		Document doc = sDocBuilder.newDocument();
		AccessRequest ar = Identifiers.createAr("AR100", null);
		
		Element res = sAccessRequestHandler.toElement(ar, doc);
		assertNotNull(res);
		assertEquals("access-request", res.getLocalName());
		assertNull(res.getPrefix());
		assertNull(res.getNamespaceURI());
		assertNotNull(res.getAttributeNode("name"));
		assertEquals("AR100", res.getAttribute("name"));
		
		assertNull(res.getAttributeNode("administrative-domain"));
		
		assertEquals(0, res.getChildNodes().getLength());
		assertEquals(1, res.getAttributes().getLength());
	}

	@Test
	public void testToElementWithNameAndAd() throws MarshalException {
		Document doc = sDocBuilder.newDocument();
		AccessRequest ar = Identifiers.createAr("AR100", "mydomain");
		
		Element res = sAccessRequestHandler.toElement(ar, doc);
		assertNotNull(res);
		assertEquals("access-request", res.getLocalName());
		assertNull(res.getPrefix());
		assertNull(res.getNamespaceURI());
		assertNotNull(res.getAttributeNode("name"));
		assertEquals("AR100", res.getAttribute("name"));
		
		assertNotNull(res.getAttributeNode("administrative-domain"));
		assertEquals("mydomain", res.getAttribute("administrative-domain"));
		
		assertEquals(0, res.getChildNodes().getLength());
		assertEquals(2, res.getAttributes().getLength());
	}

	@Test
	public void testFromElementWithName() throws UnmarshalException {
		Document doc = sDocBuilder.newDocument();
		Element xmlAr = doc.createElementNS(null, "access-request");
		xmlAr.setAttribute("name", "AR100");
		
		AccessRequest ar = sAccessRequestHandler.fromElement(xmlAr);
		assertNotNull(ar);
		assertNotNull(ar.getName());
		assertEquals("AR100", ar.getName());
		
		assertNull(ar.getAdministrativeDomain());
	}

	@Test
	public void testFromElementWithNameAndAd() throws UnmarshalException {
		Document doc = sDocBuilder.newDocument();
		Element xmlAr = doc.createElementNS(null, "access-request");
		xmlAr.setAttribute("name", "AR100");
		xmlAr.setAttribute("administrative-domain", "mydomain");
		
		AccessRequest ar = sAccessRequestHandler.fromElement(xmlAr);
		assertEquals("AR100", ar.getName());
		assertEquals("mydomain", ar.getAdministrativeDomain());
	}
	
	@Test(expected=UnmarshalException.class)
	public void testFromElementWithNullName() throws UnmarshalException {
		Document doc = sDocBuilder.newDocument();
		Element xmlAr = doc.createElementNS(null, "access-request");
		
		// throws
		AccessRequest ar = sAccessRequestHandler.fromElement(xmlAr);
		assertNull(ar);
	}

	@Test(expected=UnmarshalException.class)
	public void testFromElementWithEmptyName() throws UnmarshalException {
		Document doc = sDocBuilder.newDocument();
		Element xmlAr = doc.createElementNS(null, "access-request");
		xmlAr.setAttribute("name", "");

		// throws
		AccessRequest ar = sAccessRequestHandler.fromElement(xmlAr);
		assertNull(ar);
	}
	
	@Test
	public void testNotResponsible() throws UnmarshalException {
		Document doc = sDocBuilder.newDocument();
		Element xmlAr = doc.createElementNS(null, "ip-address");
		
		AccessRequest ar = sAccessRequestHandler.fromElement(xmlAr);
		assertNull(ar);
	}
	
	@Test(expected=MarshalException.class)
	public void testToElementWrongIdentifierType() throws UnmarshalException, MarshalException {
		Document doc = sDocBuilder.newDocument();
		Element ret = sAccessRequestHandler.toElement(Identifiers.createIp4("192.168.0.1"), doc);
		assertNull(ret);
	}
}
