package de.fhhannover.inform.trust.ifmapj.identifier;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;

import javax.xml.parsers.DocumentBuilder;

import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import de.fhhannover.inform.trust.ifmapj.exception.MarshalException;
import de.fhhannover.inform.trust.ifmapj.exception.UnmarshalException;

import util.DomHelpers;

public class IpAddressHandlerTest {
	
	private static DocumentBuilder sDocBuilder = DomHelpers.newDocumentBuilder();
	private static IpAddressHandler sIpAddressHandler = new IpAddressHandler();
	
	@Test(expected=MarshalException.class)
	public void testToElementWithNullValue() throws MarshalException {
		Document doc = sDocBuilder.newDocument();
		IpAddress ip = Identifiers.createIp4(null);
		Element res = sIpAddressHandler.toElement(ip, doc);
		assertNull(res);
	}
		
	@Test(expected=MarshalException.class)
	public void testToElementWithEmptyValue() throws MarshalException {
		Document doc = sDocBuilder.newDocument();
		IpAddress ip = Identifiers.createIp4("");
		Element res = sIpAddressHandler.toElement(ip, doc);
		assertNull(res);
	}
	
	@Test
	public void testToElementNullTypeIPv4() throws MarshalException {
		Document doc = sDocBuilder.newDocument();
		
		IpAddress ip = Identifiers.createIp(null, "192.168.0.1", "");
		
		Element res = sIpAddressHandler.toElement(ip, doc);
		assertNotNull(res);
		assertEquals("ip-address", res.getLocalName());
		assertNull(res.getPrefix());
		assertNull(res.getNamespaceURI());
		
		assertNotNull(res.getAttributeNode("value"));
		assertEquals("192.168.0.1", res.getAttribute("value"));
	
		assertNotNull(res.getAttributeNode("type"));
		assertEquals("IPv4", res.getAttribute("type"));
		
		assertNull(res.getAttributeNode("administrative-domain"));
		
		assertEquals(0, res.getChildNodes().getLength());
		assertEquals(2, res.getAttributes().getLength());
	}

	/**
	 * This shouldn't show the administrative-domain
	 * @throws MarshalException 
	 */
	@Test
	public void testToElementWithEmptyAd() throws MarshalException {
		Document doc = sDocBuilder.newDocument();
		IpAddress ip = Identifiers.createIp4("192.168.0.1", "");
		
		Element res = sIpAddressHandler.toElement(ip, doc);
		assertNotNull(res);
		assertEquals("ip-address", res.getLocalName());
		assertNull(res.getPrefix());
		assertNull(res.getNamespaceURI());
		
		assertNotNull(res.getAttributeNode("value"));
		assertEquals("192.168.0.1", res.getAttribute("value"));
	
		// default is IPv4
		assertNotNull(res.getAttributeNode("type"));
		assertEquals("IPv4", res.getAttribute("type"));
		
		assertNull(res.getAttributeNode("administrative-domain"));
		
		assertEquals(0, res.getChildNodes().getLength());
		assertEquals(2, res.getAttributes().getLength());
	}
		
	
	@Test
	public void testToElementWithValueIPv46() throws MarshalException {
		Document doc = sDocBuilder.newDocument();
		IpAddress ip = Identifiers.createIp(IpAddressType.IPv4, "192.168.0.1", null);
		
		Element res = sIpAddressHandler.toElement(ip, doc);
		assertNotNull(res);
		assertEquals("ip-address", res.getLocalName());
		assertNull(res.getPrefix());
		assertNull(res.getNamespaceURI());
		
		assertNotNull(res.getAttributeNode("value"));
		assertEquals("192.168.0.1", res.getAttribute("value"));
		
		assertNotNull(res.getAttributeNode("type"));
		assertEquals("IPv4", res.getAttribute("type"));
		
		assertNull(res.getAttributeNode("administrative-domain"));
		
		assertEquals(0, res.getChildNodes().getLength());
		assertEquals(2, res.getAttributes().getLength());

		ip = Identifiers.createIp(IpAddressType.IPv6, "192.168.0.1", null);
		// for simplicity we leave a IPv4 value, it's not checked...
		
		res = sIpAddressHandler.toElement(ip, doc);
		assertNotNull(res);
		assertEquals("ip-address", res.getLocalName());
		assertNull(res.getPrefix());
		assertNull(res.getNamespaceURI());
		
		assertNotNull(res.getAttributeNode("value"));
		assertEquals("192.168.0.1", res.getAttribute("value"));
		
		assertNotNull(res.getAttributeNode("type"));
		assertEquals("IPv6", res.getAttribute("type"));
		
		assertNull(res.getAttributeNode("administrative-domain"));
		
		assertEquals(0, res.getChildNodes().getLength());
		assertEquals(2, res.getAttributes().getLength());
	}

	@Test
	public void testToElementWithValueAndAdIPv46() throws MarshalException {
		Document doc = sDocBuilder.newDocument();
		IpAddress ip = Identifiers.createIp(IpAddressType.IPv4, "192.168.0.1",
				"mydomain");
		
		Element res = sIpAddressHandler.toElement(ip, doc);
		assertNotNull(res);
		assertEquals("ip-address", res.getLocalName());
		assertNull(res.getPrefix());
		assertNull(res.getNamespaceURI());
		
		assertNotNull(res.getAttributeNode("value"));
		assertEquals("192.168.0.1", res.getAttribute("value"));
		
		assertNotNull(res.getAttributeNode("type"));
		assertEquals("IPv4", res.getAttribute("type"));
		
		assertNotNull(res.getAttributeNode("administrative-domain"));
		assertEquals("mydomain", res.getAttribute("administrative-domain"));
		
		assertEquals(0, res.getChildNodes().getLength());
		assertEquals(3, res.getAttributes().getLength());

		// for simplicity we leave a IPv4 value, it's not checked...
		ip = Identifiers.createIp(IpAddressType.IPv6, "192.168.0.1",
				"mydomain");
		
		res = sIpAddressHandler.toElement(ip, doc);
		assertNotNull(res);
		assertEquals("ip-address", res.getLocalName());
		assertNull(res.getPrefix());
		assertNull(res.getNamespaceURI());
		
		assertNotNull(res.getAttributeNode("value"));
		assertEquals("192.168.0.1", res.getAttribute("value"));
		
		assertNotNull(res.getAttributeNode("type"));
		assertEquals("IPv6", res.getAttribute("type"));
		
		assertNotNull(res.getAttributeNode("administrative-domain"));
		assertEquals("mydomain", res.getAttribute("administrative-domain"));
		
		assertEquals(0, res.getChildNodes().getLength());
		assertEquals(3, res.getAttributes().getLength());
	}

	@Test
	public void testFromElementWithValueIPv46() throws UnmarshalException {
		Document doc = sDocBuilder.newDocument();
		Element xmlIp = doc.createElementNS(null, "ip-address");
		
		xmlIp.setAttribute("value", "192.168.0.1");
		xmlIp.setAttribute("type", "IPv4");
		
		IpAddress ip = sIpAddressHandler.fromElement(xmlIp);
		assertNotNull(ip);
		assertNotNull(ip.getValue());
		assertEquals("192.168.0.1", ip.getValue());
		assertEquals(IpAddressType.IPv4, ip.getType());
		
		assertNull(ip.getAdministrativeDomain());

		// again, leave the value IPv4, it's not checked
		xmlIp.setAttribute("type", "IPv6");
		ip = sIpAddressHandler.fromElement(xmlIp);
		assertNotNull(ip);
		assertNotNull(ip.getValue());
		assertEquals("192.168.0.1", ip.getValue());
		assertEquals(IpAddressType.IPv6, ip.getType());
		assertNull(ip.getAdministrativeDomain());

		// try no type at all, should default to IPv4
		xmlIp.removeAttribute("type");
		ip = sIpAddressHandler.fromElement(xmlIp);
		assertNotNull(ip);
		assertNotNull(ip.getValue());
		assertEquals("192.168.0.1", ip.getValue());
		assertEquals(IpAddressType.IPv4, ip.getType());
		assertNull(ip.getAdministrativeDomain());
	}

	@Test(expected=UnmarshalException.class)
	public void testFromElementWithBadType() throws UnmarshalException {
		Document doc = sDocBuilder.newDocument();
		Element xmlIp = doc.createElementNS(null, "ip-address");
		xmlIp.setAttribute("value", "192.168.0.1");
		xmlIp.setAttribute("type", "IPxx");
		IpAddress ip = sIpAddressHandler.fromElement(xmlIp);
		assertNull(ip);
	}

	@Test
	public void testFromElementWithValueAndAdIPv46() throws UnmarshalException {
		Document doc = sDocBuilder.newDocument();
		Element xmlIp = doc.createElementNS(null, "ip-address");
		
		xmlIp.setAttribute("value", "192.168.0.1");
		xmlIp.setAttribute("type", "IPv4");
		xmlIp.setAttribute("administrative-domain", "mydomain");
		
		IpAddress ip = sIpAddressHandler.fromElement(xmlIp);
		assertNotNull(ip);
		assertNotNull(ip.getValue());
		assertEquals("192.168.0.1", ip.getValue());
		assertEquals(IpAddressType.IPv4, ip.getType());
		assertEquals("mydomain", ip.getAdministrativeDomain());

		// again, leave the value IPv4, it's not checked
		xmlIp.setAttribute("type", "IPv6");
		ip = sIpAddressHandler.fromElement(xmlIp);
		assertNotNull(ip);
		assertNotNull(ip.getValue());
		assertEquals("192.168.0.1", ip.getValue());
		assertEquals(IpAddressType.IPv6, ip.getType());
		assertEquals("mydomain", ip.getAdministrativeDomain());

		// try no type at all, should default to IPv4
		xmlIp.removeAttribute("type");
		ip = sIpAddressHandler.fromElement(xmlIp);
		assertNotNull(ip);
		assertNotNull(ip.getValue());
		assertEquals("192.168.0.1", ip.getValue());
		assertEquals(IpAddressType.IPv4, ip.getType());
		assertEquals("mydomain", ip.getAdministrativeDomain());
	}

	@Test(expected=UnmarshalException.class)
	public void testFromElementNullValue() throws UnmarshalException {
		Document doc = sDocBuilder.newDocument();
		Element xmlIp = doc.createElementNS(null, "ip-address");
		IpAddress ip = sIpAddressHandler.fromElement(xmlIp);
		assertNull(ip);
	}

	@Test(expected=UnmarshalException.class)
	public void testFromElementEmptyValue() throws UnmarshalException {
		Document doc = sDocBuilder.newDocument();
		Element xmlIp = doc.createElementNS(null, "ip-address");
		xmlIp.setAttribute("value", "");
		IpAddress ip = sIpAddressHandler.fromElement(xmlIp);
		assertNull(ip);
	}
		
	@Test
	public void testNotResponsible() throws UnmarshalException {
		Document doc = sDocBuilder.newDocument();
		Element xmlIp = doc.createElementNS(null, "access-request");
		
		IpAddress ip = sIpAddressHandler.fromElement(xmlIp);
		assertNull(ip);
	}
	
	@Test(expected=MarshalException.class)
	public void testToElementWrongIdentifierType() throws UnmarshalException, MarshalException {
		Document doc = sDocBuilder.newDocument();
		Element ret = sIpAddressHandler.toElement(Identifiers.createDev("DEV100"), doc);
		assertNull(ret);
	}
}
