package de.fhhannover.inform.trust.ifmapj.identifier;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;

import javax.xml.parsers.DocumentBuilder;

import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import de.fhhannover.inform.trust.ifmapj.exception.MarshalException;
import de.fhhannover.inform.trust.ifmapj.exception.UnmarshalException;

import util.DomHelpers;

public class MacAddressHandlerTest {
	
	private static DocumentBuilder sDocBuilder = DomHelpers.newDocumentBuilder();
	private static MacAddressHandler sMacAddressHandler = new MacAddressHandler();
	
	@Test(expected=MarshalException.class)
	public void testToElementWithNullName() throws MarshalException {
		Document doc = sDocBuilder.newDocument();
		MacAddress mac = Identifiers.createMac(null);
		
		Element res = sMacAddressHandler.toElement(mac, doc);
		assertNull(res);
	}

	@Test(expected=MarshalException.class)
	public void testToElementWithNoName() throws MarshalException {
		Document doc = sDocBuilder.newDocument();
		MacAddress mac = Identifiers.createMac("");
		
		Element res = sMacAddressHandler.toElement(mac, doc);
		assertNull(res);
	}

	/**
	 * This shouldn't show the administrative-domain
	 * @throws MarshalException 
	 */
	@Test
	public void testToElementWithEmptyAd() throws MarshalException {
		Document doc = sDocBuilder.newDocument();
		MacAddress mac = Identifiers.createMac("aa:bb:cc:dd:ee:ff", "");
		
		Element res = sMacAddressHandler.toElement(mac, doc);
		assertNotNull(res);
		assertEquals("mac-address", res.getLocalName());
		assertNull(res.getPrefix());
		assertNull(res.getNamespaceURI());
		assertNotNull(res.getAttributeNode("value"));
		assertEquals("aa:bb:cc:dd:ee:ff", res.getAttribute("value"));
		
		assertNull(res.getAttributeNode("administrative-domain"));
		
		assertEquals(0, res.getChildNodes().getLength());
		assertEquals(1, res.getAttributes().getLength());
	}
		
	
	@Test
	public void testToElementWithName() throws MarshalException {
		Document doc = sDocBuilder.newDocument();
		MacAddress mac = Identifiers.createMac("aa:bb:cc:dd:ee:ff", null);
		
		Element res = sMacAddressHandler.toElement(mac, doc);
		assertNotNull(res);
		assertEquals("mac-address", res.getLocalName());
		assertNull(res.getPrefix());
		assertNull(res.getNamespaceURI());
		assertNotNull(res.getAttributeNode("value"));
		assertEquals("aa:bb:cc:dd:ee:ff", res.getAttribute("value"));
		
		assertNull(res.getAttributeNode("administrative-domain"));
		
		assertEquals(0, res.getChildNodes().getLength());
		assertEquals(1, res.getAttributes().getLength());
	}

	@Test
	public void testToElementWithNameAndAd() throws MarshalException {
		Document doc = sDocBuilder.newDocument();
		MacAddress mac = Identifiers.createMac("aa:bb:cc:dd:ee:ff", "mydomain");
		
		Element res = sMacAddressHandler.toElement(mac, doc);
		assertNotNull(res);
		assertEquals("mac-address", res.getLocalName());
		assertNull(res.getPrefix());
		assertNull(res.getNamespaceURI());
		assertNotNull(res.getAttributeNode("value"));
		assertEquals("aa:bb:cc:dd:ee:ff", res.getAttribute("value"));
		
		assertNotNull(res.getAttributeNode("administrative-domain"));
		assertEquals("mydomain", res.getAttribute("administrative-domain"));
		
		assertEquals(0, res.getChildNodes().getLength());
		assertEquals(2, res.getAttributes().getLength());
	}

	@Test
	public void testFromElementWithName() throws UnmarshalException {
		Document doc = sDocBuilder.newDocument();
		Element xmlMac = doc.createElementNS(null, "mac-address");
		xmlMac.setAttribute("value", "aa:bb:cc:dd:ee:ff");
		
		MacAddress mac = sMacAddressHandler.fromElement(xmlMac);
		assertNotNull(mac);
		assertNotNull(mac.getValue());
		assertEquals("aa:bb:cc:dd:ee:ff", mac.getValue());
		
		assertNull(mac.getAdministrativeDomain());
	}

	@Test
	public void testFromElementWithNameAndAd() throws UnmarshalException {
		Document doc = sDocBuilder.newDocument();
		Element xmlMac = doc.createElementNS(null, "mac-address");
		xmlMac.setAttribute("value", "aa:bb:cc:dd:ee:ff");
		xmlMac.setAttribute("administrative-domain", "mydomain");
		
		MacAddress mac = sMacAddressHandler.fromElement(xmlMac);
		assertEquals("aa:bb:cc:dd:ee:ff", mac.getValue());
		assertEquals("mydomain", mac.getAdministrativeDomain());
	}

	@Test(expected=UnmarshalException.class)
	public void testFromElementWithNullName() throws UnmarshalException {
		Document doc = sDocBuilder.newDocument();
		Element xmlMac = doc.createElementNS(null, "mac-address");
		MacAddress mac = sMacAddressHandler.fromElement(xmlMac);
		assertNull(mac);
	}

	@Test(expected=UnmarshalException.class)
	public void testFromElementWithEmptyName() throws UnmarshalException {
		Document doc = sDocBuilder.newDocument();
		Element xmlMac = doc.createElementNS(null, "mac-address");
		xmlMac.setAttribute("value", "");
		MacAddress mac = sMacAddressHandler.fromElement(xmlMac);
		assertNull(mac);
	}
		
	@Test
	public void testNotResponsible() throws UnmarshalException {
		Document doc = sDocBuilder.newDocument();
		Element xmlMac = doc.createElementNS(null, "ip-address");
		
		MacAddress mac = sMacAddressHandler.fromElement(xmlMac);
		assertNull(mac);
	}
	
	@Test(expected=MarshalException.class)
	public void testToElementWrongIdentifierType() throws UnmarshalException, MarshalException {
		Document doc = sDocBuilder.newDocument();
		Element ret = sMacAddressHandler.toElement(Identifiers.createDev("DEV100"), doc);
		assertNull(ret);
	}
}
