package de.fhhannover.inform.trust.ifmapj.messages;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;

import javax.xml.parsers.DocumentBuilder;

import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import util.DomHelpers;
import de.fhhannover.inform.trust.ifmapj.exception.IfmapErrorResult;
import de.fhhannover.inform.trust.ifmapj.exception.MarshalException;
import de.fhhannover.inform.trust.ifmapj.exception.UnmarshalException;

/**
 * We rely on the {@link SearchRequestHandlerTest} in order to make sure the
 * updates are correct... :-/
 */
public class SubscribeRequestHandlerTest {
	
	private static RequestHandler<? extends Request> sHandler = makeHandler();
	private static DocumentBuilder sDocBuilder = DomHelpers.newDocumentBuilder();
	private static final String REQ_EL_NAME = "subscribe";
	private static final String RES_EL_NAME = "subscribeReceived";
	private static final String IFMAP_URI = "http://www.trustedcomputinggroup.org/2010/IFMAP/2";
	
	private static SubscribeRequest makeRequest() {
		return new SubscribeRequestImpl();
	}

	private static SubscribeRequest makeSimpleRequest() {
		SubscribeRequest ret = makeRequest();
		ret.addSubscribeElement(
				TestHelpers.subscribeUpdate("mysub"));
				
		return ret;
	

	}

	private static RequestHandler<? extends Request> makeHandler() {
		return new SubscribeRequestHandler();
	}
	
	@Test
	public void testToElementGood() throws MarshalException {
		Request req = makeSimpleRequest();
		req.setSessionId("1234");
		Document doc = sDocBuilder.newDocument();
		Element ret = sHandler.toElement(req, doc);
		assertNotNull(ret);
		assertEquals(REQ_EL_NAME, ret.getLocalName());
		assertEquals(IFMAP_URI, ret.getNamespaceURI());
		assertEquals(1, ret.getAttributes().getLength());
		assertEquals("1234", ret.getAttribute("session-id"));
		assertEquals(1, ret.getChildNodes().getLength());
	}

	@Test(expected=MarshalException.class)
	public void testToElementNullSessionId() throws MarshalException {
		Request req = makeSimpleRequest();
		Document doc = sDocBuilder.newDocument();
		Element ret = sHandler.toElement(req, doc);
		assertNull(ret);
	}
	
	@Test(expected=MarshalException.class)
	public void testToElementEmptySessionId() throws MarshalException {
		Request req = makeSimpleRequest();
		req.setSessionId("");
		Document doc = sDocBuilder.newDocument();
		Element ret = sHandler.toElement(req, doc);
		assertNull(ret);
	}

	@Test(expected=MarshalException.class)
	public void testToElementWrongType() throws MarshalException {
		Request req = new SubscribeRequestImpl();
		req.setSessionId("1234");
		Document doc = sDocBuilder.newDocument();
		Element ret = sHandler.toElement(req, doc);
		assertNull(ret);
	}

	@Test(expected=MarshalException.class)
	public void testToElementNoSubscribeElements() throws MarshalException {
		Request req = makeRequest();
		req.setSessionId("1234");
		Document doc = sDocBuilder.newDocument();
		Element ret = sHandler.toElement(req, doc);
		assertNull(ret);
	}
	
	@Test(expected=MarshalException.class)
	public void testToElementEmptyNameUpdate() throws MarshalException {
		SubscribeRequest req = makeRequest();
		req.addSubscribeElement(TestHelpers.subscribeUpdate(""));
		req.setSessionId("1234");
		Document doc = sDocBuilder.newDocument();
		Element ret = sHandler.toElement(req, doc);
		assertNull(ret);
	}

	@Test(expected=MarshalException.class)
	public void testToElementNullNameUpdate() throws MarshalException {
		SubscribeRequest req = makeRequest();
		req.addSubscribeElement(TestHelpers.subscribeUpdate(null));
		req.setSessionId("1234");
		Document doc = sDocBuilder.newDocument();
		Element ret = sHandler.toElement(req, doc);
		assertNull(ret);
	}

	@Test(expected=MarshalException.class)
	public void testToElementEmptyNameDelete() throws MarshalException {
		SubscribeRequest req = makeRequest();
		req.setSessionId("1234");
		req.addSubscribeElement(TestHelpers.subscribeDelete(""));
		Document doc = sDocBuilder.newDocument();
		Element ret = sHandler.toElement(req, doc);
		assertNull(ret);
	}

	@Test(expected=MarshalException.class)
	public void testToElementNullNameDelete() throws MarshalException {
		SubscribeRequest req = makeRequest();
		req.setSessionId("1234");
		req.addSubscribeElement(TestHelpers.subscribeDelete(null));
		Document doc = sDocBuilder.newDocument();
		Element ret = sHandler.toElement(req, doc);
		assertNull(ret);
	}
	
	@Test
	public void testFromElementGood() throws UnmarshalException, IfmapErrorResult {
		Document doc = sDocBuilder.newDocument();
		Element response = doc.createElementNS(IFMAP_URI, "ifmap:response");
		Element result = doc.createElementNS(null, RES_EL_NAME);
		response.appendChild(result);
		Result res = sHandler.fromElement(response);
		assertNull(res);	// This handler returns null on success
	}
	
	@Test(expected=IfmapErrorResult.class)
	public void testFromElementWithErrorResult() throws IfmapErrorResult, UnmarshalException {
		Document doc = sDocBuilder.newDocument();
		Element response = doc.createElementNS(IFMAP_URI, "ifmap:response");
		Element result = doc.createElementNS(null, "errorResult");
		result.setAttribute("errorCode", "SystemError");
		response.appendChild(result);
		Result res = sHandler.fromElement(response);
		assertNull(res);
	}

	@Test(expected=UnmarshalException.class)
	public void testFromElementNoResult() throws IfmapErrorResult, UnmarshalException {
		Document doc = sDocBuilder.newDocument();
		Element response = doc.createElementNS(IFMAP_URI, "ifmap:response");
		Result res = sHandler.fromElement(response);
		assertNull(res);
	}

	@Test(expected=UnmarshalException.class)
	public void testFromElementWrongResult() throws IfmapErrorResult, UnmarshalException {
		Document doc = sDocBuilder.newDocument();
		Element response = doc.createElementNS(IFMAP_URI, "ifmap:response");
		Element result = doc.createElementNS(null, "publishReceived");
		response.appendChild(result);
		Result res = sHandler.fromElement(response);
		assertNull(res);
	}
	
	@Test
	public void testToElementSingleUpdate() throws MarshalException {
		SubscribeRequest req = makeRequest();
		req.setSessionId("1234");
		req.addSubscribeElement(
				TestHelpers.subscribeUpdate("mysub"));
		Document doc = sDocBuilder.newDocument();
		Element ret = sHandler.toElement(req, doc);
		assertNotNull(ret);
		assertEquals(REQ_EL_NAME, ret.getLocalName());
		assertEquals(IFMAP_URI, ret.getNamespaceURI());
		assertEquals(1, ret.getAttributes().getLength());
		assertEquals("1234", ret.getAttribute("session-id"));
		assertEquals(1, ret.getChildNodes().getLength());
		Element upEl = (Element) ret.getChildNodes().item(0);
		assertEquals("update", upEl.getLocalName());
		assertNull(upEl.getNamespaceURI());
		assertNotNull(upEl.getAttributeNode("name"));
		assertEquals("mysub", upEl.getAttribute("name"));
	}

	@Test
	public void testToElementSingleDelete() throws MarshalException {
		SubscribeRequest req = makeRequest();
		req.setSessionId("1234");
		req.addSubscribeElement(
				TestHelpers.subscribeDelete("mysub"));
		Document doc = sDocBuilder.newDocument();
		Element ret = sHandler.toElement(req, doc);
		assertNotNull(ret);
		assertEquals(REQ_EL_NAME, ret.getLocalName());
		assertEquals(IFMAP_URI, ret.getNamespaceURI());
		assertEquals(1, ret.getAttributes().getLength());
		assertEquals("1234", ret.getAttribute("session-id"));
		assertEquals(1, ret.getChildNodes().getLength());
		Element delEl = (Element) ret.getChildNodes().item(0);
		assertEquals("delete", delEl.getLocalName());
		assertNull(delEl.getNamespaceURI());
		assertNotNull(delEl.getAttributeNode("name"));
		assertEquals("mysub", delEl.getAttribute("name"));
	}
	
	@Test
	public void testToElementMix() throws MarshalException {
		SubscribeRequest req = makeRequest();
		req.setSessionId("1234");
		req.addSubscribeElement(
				TestHelpers.subscribeDelete("mysub"));
		req.addSubscribeElement(
				TestHelpers.subscribeUpdate("mysub"));
		
		Document doc = sDocBuilder.newDocument();
		Element ret = sHandler.toElement(req, doc);
		assertNotNull(ret);
		assertEquals(REQ_EL_NAME, ret.getLocalName());
		assertEquals(IFMAP_URI, ret.getNamespaceURI());
		assertEquals(1, ret.getAttributes().getLength());
		assertEquals("1234", ret.getAttribute("session-id"));
		assertEquals(2, ret.getChildNodes().getLength());
		Element delEl = (Element) ret.getChildNodes().item(0);
		Element upEl = (Element) ret.getChildNodes().item(1);
		assertEquals("delete", delEl.getLocalName());
		assertEquals("update", upEl.getLocalName());
		assertNull(upEl.getNamespaceURI());
		assertNull(delEl.getNamespaceURI());
		assertNotNull(upEl.getAttributeNode("name"));
		assertEquals("mysub", upEl.getAttribute("name"));
		assertNotNull(delEl.getAttributeNode("name"));
		assertEquals("mysub", delEl.getAttribute("name"));
	}

	@Test
	public void testToElementSingleUpdateWithFilter() throws MarshalException {
		SubscribeRequest req = makeRequest();
		req.setSessionId("1234");
		SubscribeUpdate up = (SubscribeUpdate) TestHelpers.subscribeUpdate("mysub");
		up.setResultFilter("myfilter");
		req.addSubscribeElement(up);
		Document doc = sDocBuilder.newDocument();
		Element ret = sHandler.toElement(req, doc);
		assertNotNull(ret);
		assertEquals(REQ_EL_NAME, ret.getLocalName());
		assertEquals(IFMAP_URI, ret.getNamespaceURI());
		assertEquals(1, ret.getAttributes().getLength());
		assertEquals("1234", ret.getAttribute("session-id"));
		assertEquals(1, ret.getChildNodes().getLength());
		Element upEl = (Element) ret.getChildNodes().item(0);
		assertEquals("update", upEl.getLocalName());
		assertNull(upEl.getNamespaceURI());
		assertNotNull(upEl.getAttributeNode("name"));
		assertEquals("mysub", upEl.getAttribute("name"));
		assertEquals("myfilter", upEl.getAttribute("result-filter"));
		assertNull(upEl.getAttributeNode("match-links"));
	}
}
